@@ -5,15 +5,19 @@
 /*                       Copyright (c) 1996,1997                         */
 /*                        All Rights Reserved.                           */
 /*                                                                       */
-/*  Permission to use, copy, modify, distribute this software and its    */
-/*  documentation for research, educational and individual use only, is  */
-/*  hereby granted without fee, subject to the following conditions:     */
+/*  Permission is hereby granted, free of charge, to use and distribute  */
+/*  this software and its documentation without restriction, including   */
+/*  without limitation the rights to use, copy, modify, merge, publish,  */
+/*  distribute, sublicense, and/or sell copies of this work, and to      */
+/*  permit persons to whom this work is furnished to do so, subject to   */
+/*  the following conditions:                                            */
 /*   1. The code must retain the above copyright notice, this list of    */
 /*      conditions and the following disclaimer.                         */
 /*   2. Any modifications must be clearly marked as such.                */
 /*   3. Original authors' names are not deleted.                         */
-/*  This software may not be used for commercial purposes without        */
-/*  specific prior written permission from the authors.                  */
+/*   4. The authors' names are not used to endorse or promote products   */
+/*      derived from this software without specific prior written        */
+/*      permission.                                                      */
 /*                                                                       */
 /*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        */
 /*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
@@ -41,46 +45,80 @@
 
 static int bl_match_entry(LISP entry,const EST_String &word);
 static int match_features(LISP req_feats, LISP act_feats);
-static LISP map_pos(LISP posmap, LISP pos);
-static LISP lex_map_phones(LISP entry);
 
 // doesn't need to be exact -- used it jumping back chunks when 
 // searching for entries with same head word
 #define DEFAULT_LEX_ENTRY_SIZE 40
 
-LISP lexicon_list = NIL;
+// Depth to build the index cache to
+#define CACHE_DEPTH 8
+
+static LISP lexicon_list = NIL;
+static Lexicon *current_lex = NULL;
+
+VAL_REGISTER_CLASS(lexicon,Lexicon)
+SIOD_REGISTER_CLASS(lexicon,Lexicon)
+
+Lexicon::Lexicon()
+{
+    type = lex_external; 
+    name = ""; 
+    binlexfp = NULL; 
+    posmap = NIL; 
+    gc_protect(&posmap);
+    addenda = NIL; 
+    gc_protect(&addenda);
+    index_cache = NIL; 
+    gc_protect(&index_cache);
+    matched_lexical_entries = NIL; 
+    gc_protect(&matched_lexical_entries);
+    pre_hooks = NIL;
+    gc_protect(&pre_hooks);
+    post_hooks = NIL;
+    gc_protect(&post_hooks);
+    bl_filename = ""; 
+    lts_method="";
+}
 
-Lexicon *current_lex = NULL;
 
 Lexicon::~Lexicon()
 {
     if (binlexfp != NULL)
 	fclose(binlexfp);
     gc_unprotect(&addenda);
+    gc_unprotect(&index_cache);
     gc_unprotect(&posmap);
     gc_unprotect(&matched_lexical_entries);
-    if (type == lex_internal)
-    {
-	delete internalized_entries[0];
-	delete internalized_entries;
-	delete internalized_headwords;
-    }
+    gc_unprotect(&pre_hooks);
+    gc_unprotect(&post_hooks);
+
 }
 
 LISP Lexicon::lookup(const EST_String &word, const LISP features)
 {
-    LISP entry,e,mapped;
-    
-    mapped = map_pos(posmap,features);
-    
-    if ((entry = lookup_addenda(word,mapped)) != NIL)
-	e = lex_map_phones(entry);
-    else if ((entry = lookup_complex(word,mapped)) != NIL)
-	e = lex_map_phones(entry);
-    else 
-	e = lex_map_phones(lookup_lts(word,mapped));
+    LISP entry,mapped,hooked;
+    EST_String sword;
 
-    return e;
+    if (pre_hooks != NIL)
+    {   // We could just call it and it wont do anything if NIL
+	// but this check may save some extra consing
+	hooked = apply_hooks(pre_hooks,
+			     cons(strintern(word),cons(features,NIL)));
+	sword = get_c_string(car(hooked));
+	mapped = map_pos(posmap,car(cdr(features)));
+    }
+    else
+    {
+	sword = word;
+	mapped = map_pos(posmap,features);
+    }
+
+    // Check addenda, them complex, then lts
+    if ((entry = lookup_addenda(sword,mapped)) == NIL)
+	if ((entry = lookup_complex(sword,mapped)) == NIL)
+	    entry = lookup_lts(sword,mapped);
+
+    return apply_hooks(post_hooks,entry);
 }
 
 LISP Lexicon::lookup_all(const EST_String &word)
@@ -95,21 +133,7 @@ LISP Lexicon::lookup_all(const EST_String &word)
 
     lookup_complex(word,flocons(-1));
 
-    return reverse(l_append(matched_lexical_entries,entries));
-}
-
-static LISP map_pos(LISP posmap, LISP pos)
-{
-    // Map specified features
-    LISP l;
-
-    if (consp(pos) || (pos == NIL))
-	return pos;
-
-    for (l=posmap; l != NIL; l=cdr(l))
-	if (siod_member_str(get_c_string(pos),car(car(l))) != NIL)
-	    return car(cdr(car(l)));
-    return pos;
+    return reverse(append(matched_lexical_entries,entries));
 }
 
 int Lexicon::in_lexicon(const EST_String &word,LISP features)
@@ -129,45 +153,86 @@ int Lexicon::in_lexicon(const EST_String &word,LISP features)
 EST_String Lexicon::str_lookup(const EST_String &word, const LISP features)
 {
     LISP entry = lookup(word,features);
-    char *s = siod_sprint(entry);
-    return EST_String(s);
+    return siod_sprint(entry);
 }
 
 LISP Lexicon::lookup_addenda(const EST_String &word,LISP features)
 {
     // Lookup addenda (for first match)
-    LISP l;
+    LISP l,potential=NIL;
 
-    for (l=addenda; l != 0; l=cdr(l))
-	if ((bl_match_entry(car(l),word) == 0) &&
-	    (match_features(features,(car(cdr(car(l)))))))
-	    return car(l); 
-
-    // Try again but this time ignoring features
     for (l=addenda; l != 0; l=cdr(l))
 	if (bl_match_entry(car(l),word) == 0)
-	    return car(l); 
+	{
+	    if (potential == NIL) // if nothing matches features this'll do
+		potential = car(l);
+	    if (match_features(features,(car(cdr(car(l))))))
+		return car(l); 
+	}
 
-    // not here
-    return NIL;
+    // If there isn't a complete match, a match of name is sufficient
+    return potential;
 }
 
 LISP Lexicon::lookup_complex(const EST_String &word,LISP features)
 {
     //  Lookup the word in a compile lexicon file 
+    int start,end;
 
     if (bl_filename == "")
 	return NIL;   // there isn't a compiled lexicon
 
     binlex_init();
+    int depth = 0;
+    matched_lexical_entries = NIL;
+    lex_entry_match = 0;
 
-    return bl_bsearch(word,features,blstart,blend);
+    bl_lookup_cache(index_cache,word,start,end,depth);
+
+    return bl_bsearch(word,features,start,end,depth);
+}
+
+void Lexicon::bl_lookup_cache(LISP cache, const EST_String &word, 
+			      int &start, int &end, int &depth)
+{
+    // Look up word in the index cache to get a better idea where 
+    // to look in the real file
+    
+    if (cdr(cache) == NIL)  // hit bottom
+    {
+	start = get_c_int(car(car(cache)));
+	end = get_c_int(cdr(car(cache)));
+    }
+    else if (fcompare(word,get_c_string(car(cdr(cache))),NULL) < 0)
+	bl_lookup_cache(siod_nth(2,cache),word,start,end,++depth);
+    else
+	bl_lookup_cache(siod_nth(3,cache),word,start,end,++depth);
+}
+
+void Lexicon::add_to_cache(LISP cache,
+			   const EST_String &word,
+			   int start,int mid, int end)
+{
+    if (cdr(cache) == NIL)  // hit bottom
+    {
+	LISP a,b;
+
+	a = cons(cons(flocons(start),flocons(mid)),NIL);
+	b = cons(cons(flocons(mid),flocons(end)),NIL);
+	
+	setcdr(cache,cons(strintern(word),cons(a,cons(b,NIL))));
+    }
+    else if (fcompare(word,get_c_string(car(cdr(cache))),NULL) < 0)
+	add_to_cache(siod_nth(2,cache),word,start,mid,end);
+    else
+	add_to_cache(siod_nth(3,cache),word,start,mid,end);
 }
 
 void Lexicon::binlex_init(void)
 {
     // Open the compiled lexicon file if not already open 
     char magic_number[20];
+    int end;
 
     if (binlexfp == NULL)
     {
@@ -203,7 +268,8 @@ void Lexicon::binlex_init(void)
 	}
 	blstart = ftell(binlexfp);
 	fseek(binlexfp,0L, SEEK_END);
-	blend = ftell(binlexfp);
+	end = ftell(binlexfp);
+	index_cache = cons(cons(flocons(blstart),flocons(end)),NIL);
     }
 }
 
@@ -221,26 +287,27 @@ LISP Lexicon::lookup_lts(const EST_String &word,LISP features)
     else if (lts_method == "lts_rules")
 	return lts(word,features,lts_ruleset);
     else if (lts_method == "none")
-	return cons(rintern(word),cons(NIL,cons(NIL,NIL)));
+	return cons(strintern(word),cons(NIL,cons(NIL,NIL)));
     else if (lts_method == "function")
 	return leval(cons(rintern("lex_user_unknown_word"),
-			  cons(siod_quote(rintern(word)),
-			       cons(siod_quote(features),NIL))),
+			  cons(quote(strintern(word)),
+			       cons(quote(features),NIL))),
 		     NIL);
     else
-    {
-	cerr << "LEXICON: unknown type of lts method " << lts_method << endl;
-	festival_error();
-    }
+	return leval(cons(rintern(lts_method),
+			  cons(quote(strintern(word)),
+			       cons(quote(features),NIL))),
+		     NIL);
 
     return NIL;
 }
 
-LISP Lexicon::bl_bsearch(const EST_String &word,LISP features,int start,int end)
+LISP Lexicon::bl_bsearch(const EST_String &word,LISP features,
+			 int start,int end, int depth)
 {
     // Do a binary search for word in file 
     LISP closest_entry;
-    int mid,compare;
+    int mid, compare;
 
     if (start==end)  // only happens if first item has been tested (and failed)
 	return NIL;
@@ -258,15 +325,19 @@ LISP Lexicon::bl_bsearch(const EST_String &word,LISP features,int start,int end)
 	mid = start + (end-start)/2;
     
     closest_entry = bl_find_next_entry(mid);
+    if ((depth < CACHE_DEPTH) &&
+	(end-start > 256))
+	add_to_cache(index_cache,get_c_string(car(closest_entry)),
+		     start,mid,end);
 
     compare = bl_match_entry(closest_entry,word);
 
     if (compare == 0)
 	return bl_find_actual_entry(mid,word,features);
     else if (compare < 0)  // too far
-	return bl_bsearch(word,features,start,mid);
+	return bl_bsearch(word,features,start,mid,++depth);
     else  // not far enough
-	return bl_bsearch(word,features,mid,end);
+	return bl_bsearch(word,features,mid,end,++depth);
 }
 
 LISP Lexicon::bl_find_next_entry(int pos)
@@ -284,11 +355,7 @@ LISP Lexicon::bl_find_next_entry(int pos)
 
 static int bl_match_entry(LISP entry,const EST_String &word)
 {
-    EST_String head;
-
-    head = get_c_string(car(entry));
-
-    return fcompare(word,head);
+    return fcompare(word,get_c_string(car(entry)),NULL);
 }    
 
 LISP Lexicon::bl_find_actual_entry(int pos,const EST_String &word,LISP features)
@@ -330,6 +397,8 @@ LISP Lexicon::bl_find_actual_entry(int pos,const EST_String &word,LISP features)
 		return n_entry;
 	}
 	n_entry = lreadf(binlexfp);
+	if (siod_eof(n_entry))
+	    return NIL;
     }
     
     return f_entry;
@@ -375,7 +444,6 @@ static void lex_add_lexicon(const EST_String &name, Lexicon *l)
 {
     //  Add lexicon to list of lexicons
     LISP lpair;
-    Lexicon *old_lex;
 
     lpair = siod_assoc_str(name,lexicon_list);
 
@@ -384,88 +452,19 @@ static void lex_add_lexicon(const EST_String &name, Lexicon *l)
     
     if (lpair == NIL)
     {
-	lexicon_list = cons(cons(rintern(name),
-				   cons(MKPTR(l),NIL)),
+	lexicon_list = cons(cons(strintern(name),
+				   cons(siod(l),NIL)),
 			    lexicon_list);
     }
     else
     {
 	cwarn << "lexicon " << name << " recreated" << endl;
-	old_lex = (Lexicon *)PTRVAL(car(cdr(lpair)));
-	delete old_lex;
-	PTRVAL(car(cdr(lpair))) = l;
+	setcar(cdr(lpair),siod(l));
     }
 
     return;
 }
 
-static LISP lex_map_phones(LISP entry)
-{
-    // Map phones in entry from lexical set to current set 
-    LISP s,p,nphones;
-    EST_String curset;
-
-    const EST_String &lexset = lex_current_phoneset();
-    curset = get_c_string(ft_get_param("PhoneSet"));
-
-    if (lexset == curset)
-	return entry;
-
-    for (s=car(cdr(cdr(entry))); s != NIL; s=cdr(s))
-    {
-	for (p=car(car(s)),nphones=NIL; p != NIL; p=cdr(p))
-	    nphones = cons(rintern(map_phone(get_c_string(car(p)),
-					     lexset,curset)),
-			   nphones);
-	CAR(car(s)) = reverse(nphones);
-    }
-    return entry;
-}
-
-void Lexicon::internalize(void)
-{
-    // Read in the full compiled lexicon. Useful when doing intensive 
-    // lexical access 
-    char *all_entries;
-    int i,j;
-
-    cerr << "Lexicon: internalize not supported" << endl;
-    festival_error();
-
-    // this code has never been run !!!!!
-
-    binlex_init();   /* ensure we have a compiled lexicon to suck in */
-    
-    if (comp_num_entries == -1)
-    {
-	cerr << "Lexicon: can't internalize old format lexicons, recompile it"
-	    << endl;
-	festival_error();
-    }
-    internalized_entries = new char *[comp_num_entries];
-    all_entries = new char[blend-blstart+1];
-    internalized_headwords = new EST_String[comp_num_entries];
-    fseek(binlexfp,blstart,SEEK_SET);
-    fread(all_entries,sizeof(char),blend-blstart,binlexfp);
-    internalized_entries[0] = all_entries;
-    for (j=1,i=0; i<blend-blstart; i++)
-	if (all_entries[i] == '\n')
-	{
-	    all_entries[i] = '\0';
-	    internalized_entries[j] = &all_entries[i+1];
-	    j++;
-	    if (j >= comp_num_entries)
-	    {
-		cerr << "Lexicon: corrupted lexicon " <<
-		    "number of entries doesn't match" << endl;
-		festival_error();
-	    }
-	}
-    all_entries[blend-blstart+1] = '\0';  // just to be safe
-    type = lex_internal;
-
-}
-
 //
 //  This functions provide the LISP access to the lexicons including
 //  access and selection
@@ -545,6 +544,28 @@ static LISP lex_set_phoneset(LISP psname)
     return psname;
 }
 
+static LISP lex_set_pre_hooks(LISP hooks)
+{
+    LISP last_hooks;
+    check_current_lex();
+
+    last_hooks = current_lex->pre_hooks;
+    current_lex->pre_hooks = hooks;
+
+    return last_hooks;
+}
+
+static LISP lex_set_post_hooks(LISP hooks)
+{
+    LISP last_hooks;
+    check_current_lex();
+
+    last_hooks = current_lex->post_hooks;
+    current_lex->post_hooks = hooks;
+
+    return last_hooks;
+}
+
 EST_String lex_current_phoneset(void)
 {
     check_current_lex();
@@ -587,7 +608,7 @@ LISP lex_select_lex(LISP lexname)
 	festival_error();
     }
     else
-	current_lex = (Lexicon *)PTRVAL(car(cdr(lpair)));
+	current_lex = lexicon(car(cdr(lpair)));
 
     return lastname;
 }
@@ -606,54 +627,47 @@ LISP lex_list(void)
 
 static LISP lex_lookup_lisp(LISP lword, LISP features)
 {
-    EST_String word = get_c_string(lword);
-
-    check_current_lex();
-
-    return current_lex->lookup(word,features);
-
+    return lex_lookup_word(get_c_string(lword),features);
 }
 
 static LISP lex_lookup_all(LISP lword)
 {
-    EST_String word = get_c_string(lword);
-
     check_current_lex();
 
-    return current_lex->lookup_all(word);
+    return current_lex->lookup_all(get_c_string(lword));
 }
 
 static LISP lex_entrycount(LISP lword)
 {
-    EST_String word = get_c_string(lword);
-
     check_current_lex();
 
     // assumes -1 is never in the pos field
-    current_lex->lookup(word,flocons(-1));
+    current_lex->lookup(get_c_string(lword),flocons(-1));
 
     return flocons(current_lex->num_matches());
 }
 
 LISP lex_lookup_word(const EST_String &word, LISP features)
 {
-
     check_current_lex();
 
     return current_lex->lookup(word,features);
-
 }
 
-int in_current_lexicon(const EST_String &word,LISP features)
+int in_current_lexicon(const EST_String &word, LISP features)
 {
     check_current_lex();
     return current_lex->in_lexicon(word,features);
 }
 
+void festival_lex_ff_init(void);
+
 void festival_Lexicon_init(void)
 {
     // define lexicon related functions
 
+    festival_lex_ff_init();
+
     init_subr_1("lex.set.compile.file",lex_set_compile_file,
     "(lex.set.compile.file COMPFILENAME)\n\
   Set the current lexicon's compile file to COMPFILENAME.  COMPFILENAME\n\
@@ -680,6 +694,14 @@ void festival_Lexicon_init(void)
     "(lex.set.pos.map POSMAP)\n\
   A reverse assoc-list mapping part of speech tags to the lexical\n\
   part of speech tag set. [see Lexical entries]");
+    init_subr_1("lex.set.pre_hooks",lex_set_pre_hooks,
+    "(lex.set.pre_hooks HOOKS)\n\
+  Set a function or list of functions that are to be applied to the entry\n\
+  before lookup.  Returns previous value [see Lexical entries]");
+    init_subr_1("lex.set.post_hooks",lex_set_post_hooks,
+    "(lex.set.post_hooks HOOKS)\n\
+  Set a function or list of functions that are to be applied to the entry\n\
+  after lookup.  Returns previous value [see Lexical entries]");
     init_subr_1("lex.set.phoneset",lex_set_phoneset,
     "(lex.set.phoneset PHONESETNAME)\n\
   Set current lexicon's phone set to PHONESETNAME.  PHONESETNAME must be\n\
@@ -742,67 +764,4 @@ void festival_Lexicon_init(void)
     "(lts.list)\n\
   Return list of all current defined LTS rulesets.");
 
-    festival_def_ff("gpos","Word",ff_word_gpos,
-    "Word.gpos\n\
-  Returns a guess at the part of speech of this word.  The lisp a-list\n\
-  guess_pos is used to load up this word.  If no part of speech is\n\
-  found in there \"content\" is returned.  This allows a quick efficient\n\
-  method for part of speech tagging into closed class and content words.");
-    festival_def_ff("cf","Word",ff_word_cf,
-    "Word.cf\n\
-  Returns 1 if this word is a content word and the next of a function\n\
-  word.  Function/content distinction is defined by the feature function\n\
-  gpos.");
-    festival_def_ff("contentp","Word",ff_word_contentp,
-    "Word.contentp\n\
-  Returns 1 if this word is a content word as defined by gpos, 0 otherwise.");
-    festival_def_ff("cap","Word",ff_word_cap,
-    "Word.cap\n\
-  Returns 1 if this word starts with a capital letter, 0 otherwise.");
-    festival_def_ff("n_content","Word",ff_word_n_content,
-    "Word.n_content\n\
-  Next content word.  Note this doesn't use the standard n. notation as\n\
-  it may have to search a number of words forward before finding a\n\
-  non-function word.  Uses gpos to define content/function word distinction.\n\
-  This also works for Tokens.");
-    festival_def_ff("nn_content","Word",ff_word_nn_content,
-    "Word.nn_content\n\
-  Next next content word.  Note this doesn't use the standard n.n. notation\n\
-  as it may have to search a number of words forward before finding the \n\
-  second non-function word.  Uses gpos to define content/function word\n\
-  distinction.  This also works for Tokens.");
-    festival_def_ff("p_content","Word",ff_word_p_content,
-    "Word.p_content\n\
-  Previous content word.  Note this doesn't use the standard p. notation\n\
-  as it may have to search a number of words backward before finding the \n\
-  first non-function word.  Uses gpos to define content/function word\n\
-  distinction.  This also works for Tokens.");
-    festival_def_ff("pp_content","Word",ff_word_pp_content,
-    "Word.pp_content\n\
-  Previous previous content word.  Note this doesn't use the standard p.p.\n\
-  notation as it may have to search a number of words backward before\n\
-  finding the first non-function word.  Uses gpos to define \n\
-  content/function word distinction.  This also works for Tokens.");
-    festival_def_ff("syl_reduced","Syllable",ff_syl_reduced,
-    "Syllable.syl_reduced\n\
-  This checks the actual vowel in a syllable against the vowel in the\n\
-  corresponding syllable in the lexical entry of the word this syllable\n\
-  is contained within.  This is used in training vowel reduction modules.\n\
-  If the actual vowel is a reduced form of the lexical vowel with respect\n\
-  to postlex_vowel_reduce_table 1 is returned.  If the vowels match\n\
-  0 is returned.  If they don't match and are not in the vowel reduce table\n\
-  0 is also returned but the mis-match is printed so that you may add the\n\
-  reduction to the table.");
-    festival_def_ff("syl_lex_vowel","Syllable",ff_syl_lex_vowel,
-    "Syllable.syl_lex_vowel\n\
-  The vowel in the syllable in the lexical entry corrsponding to this\n\
-  syllable.  Will return unknown when one can't be determined.");
-    festival_def_ff("syl_onset_type","Syllable",ff_syl_onset_type,
-    "Syllable.syl_onset_type\n\
-  Return the van Santen and Hirschberg classification. -V for unvoiced,\n\
-  +V-S for voiced but no sonorants, and +S for sonorants.");
-    festival_def_ff("syl_coda_type","Syllable",ff_syl_coda_type,
-    "Syllable.syl_coda_type\n\
-  Return the van Santen and Hirschberg classification. -V for unvoiced,\n\
-  +V-S for voiced but no sonorants, and +S for sonorants.");
 }
