@@ -5,15 +5,19 @@
 /*                       Copyright (c) 1996,1997                         */
 /*                        All Rights Reserved.                           */
 /*                                                                       */
-/*  Permission to use, copy, modify, distribute this software and its    */
-/*  documentation for research, educational and individual use only, is  */
-/*  hereby granted without fee, subject to the following conditions:     */
+/*  Permission is hereby granted, free of charge, to use and distribute  */
+/*  this software and its documentation without restriction, including   */
+/*  without limitation the rights to use, copy, modify, merge, publish,  */
+/*  distribute, sublicense, and/or sell copies of this work, and to      */
+/*  permit persons to whom this work is furnished to do so, subject to   */
+/*  the following conditions:                                            */
 /*   1. The code must retain the above copyright notice, this list of    */
 /*      conditions and the following disclaimer.                         */
 /*   2. Any modifications must be clearly marked as such.                */
 /*   3. Original authors' names are not deleted.                         */
-/*  This software may not be used for commercial purposes without        */
-/*  specific prior written permission from the authors.                  */
+/*   4. The authors' names are not used to endorse or promote products   */
+/*      derived from this software without specific prior written        */
+/*      permission.                                                      */
 /*                                                                       */
 /*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        */
 /*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
@@ -44,6 +48,7 @@ static unsigned int DIPHONE_MAGIC=0x46544449;  /* FTDI */
 static void load_index(DIPHONE_DATABASE *database);
 static void load_diphs(DIPHONE_DATABASE *database);
 static void load_lpc_file(DIPHONE_DATABASE *db,int diph,int mode);
+static void lpc2ref(const float *lpc, float *rfc, int order);
 static void extract_lpc_frames(DIPHONE_DATABASE *db, int diph, EST_Track &lpc);
 static void load_signal_file(DIPHONE_DATABASE *db, int i, int mode);
 static void database_malloc(int ndiphs, DIPHONE_DATABASE *database);
@@ -163,6 +168,13 @@ static void load_lpc_file(DIPHONE_DATABASE *db,int diph,int mode)
 		lpc_file << endl;
 	    festival_error();
 	}
+	if (lpc.num_channels() != db->lpc_order)
+	{
+	    cerr << "Diphone: lpc file " <<
+		lpc_file << " has order " << lpc.num_channels() << 
+		    " while database has " << db->lpc_order << endl;
+	    festival_error();
+	}
 	// Extract frames (pitch synchronously)
 	extract_lpc_frames(db,diph,lpc);
     }
@@ -170,30 +182,56 @@ static void load_lpc_file(DIPHONE_DATABASE *db,int diph,int mode)
     return;
 }
 
+static void ref2lpc(const float *rfc, float *lpc, int order)
+{
+    // Here we use Christopher Longet Higgin's algorithm converted to 
+    // an equivalent by awb.  Its doesn't have hte reverse order or
+    // negation requirement.
+    float a,b;
+    int n,k;
+    
+    for (n=0; n < order; n++)
+    {
+	lpc[n] = rfc[n];
+	for (k=0; 2*(k+1) <= n+1; k++)
+	{
+	    a = lpc[k];
+	    b = lpc[n-(k+1)];
+	    lpc[k] = a-b*lpc[n];
+	    lpc[n-(k+1)] = b-a*lpc[n];
+	}
+    }
+}
+
 static void extract_lpc_frames(DIPHONE_DATABASE *db, int diph, EST_Track &lpc)
 {
     // Extract LPC frames from lpc, one for each pitch mark
     int frame_num;
     float pos,factor;
-    int i,j;
+    float ps_pos;
+    int i,j,k;
 
     db->lpc[diph]->f = walloc(float *,db->pm[diph]->nmark);
+    float *lpcs = walloc(float,lpc.num_channels());
     for (i=0; i < db->pm[diph]->nmark; i++)
     {
 	if (db->lpc_pitch_synch)
 	{
-	    db->lpc[diph]->f[i] = walloc(float,lpc.num_channels()-1);
+	    db->lpc[diph]->f[i] = walloc(float,lpc.num_channels());
 	    pos = (((float)db->pm[diph]->mark[i]-db->sig_band)/
 		   (float)db->samp_freq)
 		+ (db->indx[diph]->beg/1000.0);
-	    pos *= 1000;
-	    for (j=0; j<lpc.num_frames(); j++)
+	    for (j=1,ps_pos=0; j<lpc.num_frames(); j++)
 	    {
-		if (fabs(pos-lpc(j,0)) < 2.0)
+		ps_pos = lpc.t(j);
+		if (fabs(pos-ps_pos) < 0.003)  // within 3 ms
 		{
-		    memmove(db->lpc[diph]->f[i],
-			    &lpc(j,1),
-			    sizeof(float)*lpc.num_channels()-1);
+		    for (k=0; k < lpc.num_channels(); k++)
+			lpcs[k] = lpc(j,k);
+		    // need to get reflection coefficients
+		    db->lpc[diph]->f[i][0] = lpcs[0];
+		    lpc2ref(&lpcs[1],&db->lpc[diph]->f[i][1],
+			    lpc.num_channels()-1);
 		    break;
 		}
 	    }
@@ -220,10 +258,12 @@ static void extract_lpc_frames(DIPHONE_DATABASE *db, int diph, EST_Track &lpc)
 		factor = (pos - ((1+frame_num)*db->lpc_frame_shift))/
 		    db->lpc_frame_shift;
 		for (j=0; j < lpc.num_channels(); j++)
+		{
 		    db->lpc[diph]->f[i][j] = 
 			lpc(frame_num,j) +
 			    (factor * (lpc(frame_num+1,j)-
 				       lpc(frame_num,j)));
+		}
 	    }
 	    if (frame_num >= lpc.num_frames())
 	    {
@@ -232,22 +272,62 @@ static void extract_lpc_frames(DIPHONE_DATABASE *db, int diph, EST_Track &lpc)
 		memset(db->lpc[diph]->f[i],0,sizeof(float)*lpc.num_channels());
 	    }
 	    else  // Last one so just take it as is
-		memmove(db->lpc[diph]->f[i],&lpc(frame_num),
-			sizeof(float)*lpc.num_channels());
+	      {
+		lpc.copy_frame_out(frame_num, db->lpc[diph]->f[i], 
+				   0, lpc.num_channels());
+	      }
 	}
     }
     db->lpc[diph]->nframes = db->pm[diph]->nmark;
-    db->lpc_order = lpc.num_channels();
+    wfree(lpcs);
+//    db->lpc_order = lpc.num_channels();
 
 }
 
+static void lpc2ref(const float *lpc, float *rfc, int order)
+{
+    // LPC to reflection coefficients 
+    // from code from Borja Etxebarria
+    int i,j;
+    float f,ai;
+    float *vo,*vx;
+    float *vn = new float[order];
+
+    i = order-1;
+    rfc[i] = ai = lpc[i];
+    f = 1-ai*ai;
+    i--;
+
+    for (j=0; j<=i; j++)
+	rfc[j] = (lpc[j]+((ai*lpc[i-j])))/f;
+
+    /* vn=vtmp in previous #define */
+    vo=rfc;
+
+    for ( ;i>0; ) 
+    {
+	ai=vo[i];
+	f = 1-ai*ai;
+	i--;
+	for (j=0; j<=i; j++)
+	    vn[j] = (vo[j]+((ai*vo[i-j])))/f;
+	
+	rfc[i]=vn[i];
+
+	vx = vn;
+	vn = vo;
+	vo = vx;
+    }
+
+    delete [] vn;
+}
+
 static void load_signal_file(DIPHONE_DATABASE *db, int i, int mode)
 {
     //  Load signal (or lpc residual) file)
     int beg_samp,end_samp,zone,nsamples;
     EST_String signal_file;
     EST_String sig_type;
-    EST_Wave w;
     int offset,error;
     beg_samp = 0;
     zone = 0;
@@ -270,6 +350,7 @@ static void load_signal_file(DIPHONE_DATABASE *db, int i, int mode)
     {
 	if (db->gtype == di_ungrouped)
 	{   
+	    EST_Wave w;
 	    if (db->type == di_lpc)
 	    {
 		signal_file = EST_String(db->lpc_dir) +
@@ -304,9 +385,8 @@ static void load_signal_file(DIPHONE_DATABASE *db, int i, int mode)
 	    else
 		error = 0;
 	    memset(db->vox[i]->signal,0,error*sizeof(short));
-	    memmove(&db->vox[i]->signal[error],
-		    w.data(),
-		    sizeof(short)*(w.num_samples()-error));
+	    for (int j=0; j < w.num_samples()-error; j++)
+		db->vox[i]->signal[error+j] = w(j);
 	    db->vox[i]->nsamples = w.num_samples()-error;
 	}
 	else //  grouped so have to access the group file
@@ -345,6 +425,102 @@ static void load_signal_file(DIPHONE_DATABASE *db, int i, int mode)
 
 }
 
+void load_pitch_file(DIPHONE_DATABASE *database, int i, int mode)
+{
+    // load files from newer Track format
+    int mark[5000];
+    EST_String pitch_file;
+    EST_Track pms;
+    float fnum;
+    int k,k1,k2,m,zone,beg_samp,p;
+
+    if ((database->pm[i]->mark != 0) ||
+	(mode != di_direct))
+	return;
+
+    pitch_file = EST_String(database->pitch_dir)+database->indx[i]->file+
+	database->pitch_ext;
+    if (pms.load(pitch_file) != format_ok)
+    {
+	cerr << "Diphone: Can't open pitch file " << pitch_file << endl;
+	festival_error();
+    }
+    /* assumptions.. only those within the limits of the diphone  */
+
+    beg_samp = (int)((database->indx[i]->beg)/1000.0*database->samp_freq);
+ 
+    zone = database->sig_band;
+
+    k = 0;
+    k1 = 0;
+    k2 = 0;
+    for (p=0; p < pms.num_frames(); p++)
+    {
+	fnum = pms.t(p)*1000.0;
+	if((fnum>database->indx[i]->beg) && (fnum<database->indx[i]->mid)) 
+	{
+	    mark[k] = (int)(fnum/1000.0*database->samp_freq - beg_samp + zone);
+	    if ((mark[k] >= database->vox[i]->nsamples+zone) ||
+		(mark[k] > 64534))
+	    {
+		fprintf(stderr,"Diphone: Mark out of range -- too large %s\n",
+			(const char *)pitch_file);
+		k--; k1--;
+	    }
+	    if(mark[k] < zone) 
+	    {
+		fprintf(stderr,"Diphone: Mark out of range -- too small %s\n",
+			(const char *)pitch_file);
+		k--; k1--;
+	    }
+	    k++;
+	    k1++;
+	} 
+	else if((fnum>=database->indx[i]->mid) && 
+		(fnum<database->indx[i]->end)) 
+	{
+	    mark[k] = (int)(fnum/1000.0*database->samp_freq - beg_samp + zone);
+	    if ((mark[k] >= database->vox[i]->nsamples+zone) ||
+		(mark[k] > 64534))
+	    {
+		fprintf(stderr,"Diphone: Mark out of range -- too large %s\n",
+			(const char *)pitch_file);
+		k--; k2--;
+	    }
+	    if(mark[k] < zone) 
+	    {
+		fprintf(stderr,"Diphone: Mark out of range -- too small %s\n",
+			(const char *)pitch_file);
+		k--; k2--;
+	    }
+	    k++;
+	    k2++;
+	}
+    }
+    database->pm[i]->mark = walloc(unsigned short,k);
+    for(m=0;m<k;m++) 
+	database->pm[i]->mark[m] = (unsigned short)mark[m];
+
+    database->pm[i]->nmark = (unsigned short)k;
+    database->pm[i]->lmark = (unsigned short)k1;
+    database->pm[i]->rmark = (unsigned short)k2;
+    if (database->pm[i]->rmark == 0)
+    {
+	*cdebug << "Diphone: modifying edge pms for " 
+	    << database->indx[i]->diph << endl;
+	database->pm[i]->rmark = 1;
+	database->pm[i]->lmark -= 1;
+    }
+    if (database->pm[i]->nmark <= 0)
+    {
+	cerr << "Diphone: diphone " << database->indx[i]->diph << 
+          " has 0 pitchmarks" << endl;
+	festival_error();
+    }
+
+}
+
+#if 0
 void load_pitch_file(DIPHONE_DATABASE *database, int i, int mode)
 {
     char s[100];
@@ -431,10 +607,16 @@ void load_pitch_file(DIPHONE_DATABASE *database, int i, int mode)
 	database->pm[i]->rmark = 1;
 	database->pm[i]->lmark -= 1;
     }
+    if (database->pm[i]->nmark <= 0)
+    {
+	cerr << "Diphone: diphone " << database->indx[i]->diph << 
+          " has 0 pitchmarks" << endl;
+	festival_error();
+    }
 
     fclose(pfd);
 }
-
+#endif
 
 /* Buffer to hold current diphone signal when using ondemand access   */
 /* method.  It remembers the last phone accessed as typical access is */
@@ -705,16 +887,21 @@ static void di_group_load_lpc_params(DIPHONE_DATABASE *db)
 		    &db->allframes[(frame_offset+j)*db->lpc_order];
 	else if (db->group_encoding == di_ulaw)
 	{
+	    int fixedpoint = FALSE;
+	    if (siod_get_lval("lpc_fixedpoint",NULL) != NIL)
+		fixedpoint = TRUE;
 	    for (j=0;j<db->lpc[i]->nframes;j++)
 	    {
 		db->lpc[i]->f[j] = walloc(float,db->lpc_order);
 		this_frame = (frame_offset+j)*db->lpc_order;
-		for (k=0;k<db->lpc_order;k++)
-		{
-		    db->lpc[i]->f[j][k] = 
-			(float)db->allframesshort[this_frame+k]/
-			    32766.0;
-		}
+		if (fixedpoint)
+		    for (k=0;k<db->lpc_order;k++)
+			db->lpc[i]->f[j][k] = 
+			    (float)db->allframesshort[this_frame+k];
+		else
+		    for (k=0;k<db->lpc_order;k++)
+			db->lpc[i]->f[j][k] = 
+			    (float)db->allframesshort[this_frame+k]/32766.0;
 	    }
 	}
 	else
@@ -798,7 +985,7 @@ void di_save_grouped_db(const EST_Pathname &filename, DIPHONE_DATABASE *db)
     LISP params;
     int strsize,totsamples,totframes;
     int i,j,k;
-    
+
     if ((fd=fopen(filename,"wb")) == NULL)
     {
 	cerr << "Diphone: cannot open group file " <<
