@@ -5,15 +5,19 @@
 /*                       Copyright (c) 1996,1997                         */
 /*                        All Rights Reserved.                           */
 /*                                                                       */
-/*  Permission to use, copy, modify, distribute this software and its    */
-/*  documentation for research, educational and individual use only, is  */
-/*  hereby granted without fee, subject to the following conditions:     */
+/*  Permission is hereby granted, free of charge, to use and distribute  */
+/*  this software and its documentation without restriction, including   */
+/*  without limitation the rights to use, copy, modify, merge, publish,  */
+/*  distribute, sublicense, and/or sell copies of this work, and to      */
+/*  permit persons to whom this work is furnished to do so, subject to   */
+/*  the following conditions:                                            */
 /*   1. The code must retain the above copyright notice, this list of    */
 /*      conditions and the following disclaimer.                         */
 /*   2. Any modifications must be clearly marked as such.                */
 /*   3. Original authors' names are not deleted.                         */
-/*  This software may not be used for commercial purposes without        */
-/*  specific prior written permission from the authors.                  */
+/*   4. The authors' names are not used to endorse or promote products   */
+/*      derived from this software without specific prior written        */
+/*      permission.                                                      */
 /*                                                                       */
 /*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        */
 /*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
@@ -76,6 +80,9 @@ static LISP diphone_dbs = NIL;
 static DIPHONE_DATABASE *di_db=0;
 static EST_Regex RXdi_ud("[$_]");
 
+VAL_REGISTER_TYPE(diphone_db,DD_STRUCT)
+SIOD_REGISTER_CLASS(diphone_db,DD_STRUCT)
+
 LISP FT_Diphone_Load_Diphones(LISP params)
 {
     DIPHONE_DATABASE *db;
@@ -106,13 +113,14 @@ void di_general_parameters(DIPHONE_DATABASE *db,LISP params)
 {
     // These may be reset from those set in group file
 
-    db->name = get_param_str("name",params,"none");
-    db->default_diphone = get_param_str("default_diphone",params,"");
+    db->name = wstrdup(get_param_str("name",params,"none"));
+    db->default_diphone = wstrdup(get_param_str("default_diphone",params,""));
     db->alternates_before = get_param_lisp("alternates_before",params,NIL);
     gc_protect(&db->alternates_before);
     db->alternates_after = get_param_lisp("alternates_after",params,NIL);
     gc_protect(&db->alternates_after);
-    db->sig_access_type_str = get_param_str("access_type",params,"direct");
+    db->sig_access_type_str = 
+	wstrdup(get_param_str("access_type",params,"direct"));
     if (streq(db->sig_access_type_str,"direct"))
 	db->sig_access_type = di_direct;
     else if (streq(db->sig_access_type_str,"dynamic"))
@@ -133,8 +141,8 @@ void di_fixed_parameters(DIPHONE_DATABASE *db,LISP params)
     // These cannot be changed in a compiled group file
     int nphones;
 
-    db->index_file = get_param_str("index_file",params,"index");
-    db->type_str = get_param_str("type",params,"pcm");
+    db->index_file = wstrdup(get_param_str("index_file",params,"index"));
+    db->type_str = wstrdup(get_param_str("type",params,"pcm"));
     if (streq(db->type_str,"pcm"))
 	db->type = di_pcm;
     else if (streq(db->type_str,"lpc"))
@@ -145,14 +153,14 @@ void di_fixed_parameters(DIPHONE_DATABASE *db,LISP params)
 	    db->type_str << "\"" << endl;
 	festival_error();
     }
-    db->signal_dir = get_param_str("signal_dir",params,"signal/");
-    db->signal_ext = get_param_str("signal_ext",params,".vox");
-    db->signal_type = get_param_str("signal_type",params,"vox");
-    db->lpc_dir = get_param_str("lpc_dir",params,"lpc/");
-    db->lpc_ext = get_param_str("lpc_ext",params,".lpc");
-    db->lpc_res_ext = get_param_str("lpc_res_ext",params,".res");
-    db->lpc_res_type = get_param_str("lpc_res_type",params,"esps");
-    db->lpc_type = get_param_str("lpc_type",params,"htk");
+    db->signal_dir = wstrdup(get_param_str("signal_dir",params,"signal/"));
+    db->signal_ext = wstrdup(get_param_str("signal_ext",params,".vox"));
+    db->signal_type = wstrdup(get_param_str("signal_type",params,"vox"));
+    db->lpc_dir = wstrdup(get_param_str("lpc_dir",params,"lpc/"));
+    db->lpc_ext = wstrdup(get_param_str("lpc_ext",params,".lpc"));
+    db->lpc_res_ext = wstrdup(get_param_str("lpc_res_ext",params,".res"));
+    db->lpc_res_type = wstrdup(get_param_str("lpc_res_type",params,"esps"));
+    db->lpc_type = wstrdup(get_param_str("lpc_type",params,"htk"));
     db->lpc_order = get_param_int("lpc_order",params,18);
     db->lpc_frame_shift = get_param_float("lpc_frame_shift",params,0.010);
     db->lpc_res_offset = get_param_float("lpc_res_offset",params,
@@ -160,13 +168,13 @@ void di_fixed_parameters(DIPHONE_DATABASE *db,LISP params)
     db->lpc_frame_offset = get_param_int("lpc_frame_offset",params,1);
     if (streq("t",get_param_str("lpc_pitch_synch",params,"f")))
 	db->lpc_pitch_synch = TRUE;
-    db->pitch_dir = get_param_str("pitch_dir",params,"pitch/");
-    db->pitch_ext = get_param_str("pitch_ext",params,".pm");
+    db->pitch_dir = wstrdup(get_param_str("pitch_dir",params,"pitch/"));
+    db->pitch_ext = wstrdup(get_param_str("pitch_ext",params,".pm"));
     db->samp_freq = get_param_int("samp_freq",params,16000);
     // old name was "group_type", new name "group_encoding"
-    db->group_encoding_str = get_param_str("group_type",params,"raw"); 
-    db->group_encoding_str = get_param_str("group_encoding",params,
-					   db->group_encoding_str);
+    db->group_encoding_str=wstrdup(get_param_str("group_type",params,"raw")); 
+    db->group_encoding_str = wstrdup(get_param_str("group_encoding",params,
+					   db->group_encoding_str));
     db->def_f0 = get_param_int("def_f0",params,125);
     if (streq(db->group_encoding_str,"raw"))
 	db->group_encoding = di_raw;
@@ -177,7 +185,7 @@ void di_fixed_parameters(DIPHONE_DATABASE *db,LISP params)
 	cerr << "Diphone: unknown group encoding" << endl;
 	festival_error();
     }
-    db->phoneset = get_param_str("phoneset",params,"none");
+    db->phoneset = wstrdup(get_param_str("phoneset",params,"none"));
     nphones = phoneset_name_to_set(db->phoneset)->num_phones();
     db->ndiphs = nphones*nphones;  // max limit of diphones
     db->ndiphs = get_param_int("num_diphones",params,db->ndiphs);
@@ -278,17 +286,13 @@ DIPHONE_DATABASE *make_diphone_db(void)
     return db;
 }
 
-// This one will be added to the EST_Wave class once I'm happy with it
-static void gc_wave(void *w) { delete (EST_Wave *)w; }
-
 LISP FT_Diphone_Synthesize_Utt(LISP utt)
 {
-    EST_Utterance *u = GETUTTVAL(utt);
+    EST_Utterance *u = get_c_utt(utt);
     DIPHONE_SPN *ps;
     DIPHONE_ACOUSTIC *as;
     DIPHONE_OUTPUT *output;
-    EST_Stream_Item item;
-    EST_Wave *w = new EST_Wave;
+    EST_Item *item=0;
 
     *cdebug << "Diphone module" << endl;
 
@@ -304,8 +308,8 @@ LISP FT_Diphone_Synthesize_Utt(LISP utt)
     /* Build structure */
     merge_silences(*u);
     ps = make_spn(*u);   // get info from utterance structure
-    as = make_as(ps);
-    output = make_output(di_db,ps->cum_dur[ps->p_sz-1]);
+
+    output = make_output(di_db,(ps->p_sz != 0 ? ps->cum_dur[ps->p_sz-1] : 0));
 
     if (ps->p_sz < 2)
     {   //  just build a silence of the specified time
@@ -313,13 +317,12 @@ LISP FT_Diphone_Synthesize_Utt(LISP utt)
     }
     else
     {
+	as = make_as(ps);
 	di_diphones(ps,di_db);
 	di_calc_pitch(di_db,ps,as);
 	di_frame_select(di_db,ps,as);
 	if (di_db->type == di_lpc)
 	    di_reslpc(di_db,as,output);       // residual excited LPC
-	else if (di_db->type == di_sine)
-	    di_sinusoid(di_db,as,output);
 	else if (di_db->type == di_pcm)
 	    di_psola_tm(di_db,as,output);     // can't be distributed
 	else
@@ -327,35 +330,32 @@ LISP FT_Diphone_Synthesize_Utt(LISP utt)
 	    cerr << "Diphone: unsupported database form\n";
 	    festival_error();
 	}
+	delete_as(as);
     }
 
     delete_spn(ps);
-    delete_as(as);
 
-    // Add wave as stream into utterance
-    w->set_data_copy(output->track,
-		     output->o_sz,
-		     di_db->samp_freq,
-		     1);
+    // Add wave into utterance
+    EST_Wave *w = new EST_Wave;
+    w->resize(output->o_sz,1,0);
+    w->set_sample_rate(di_db->samp_freq);
+    for (int i=0; i<w->length(); i++)
+	w->a_no_check(i) = output->track[i];
 
-    u->create_stream("Wave");
+    item = u->create_relation("Wave")->append();
+    item->set_val("wave",est_val(w));
 
     delete_output(output);
 
-    item.init("Wave");
-    item.set_contents(w,gc_wave);
-    u->stream("Wave").append(item);
-
     return utt;
 }
 
 LISP FT_reslpc_resynth(LISP file)
 {
     EST_Utterance *u = new EST_Utterance;
+    EST_Item *item = 0;
     DIPHONE_ACOUSTIC *as;
     DIPHONE_OUTPUT *output;
-    EST_Stream_Item item;
-    EST_Wave *w = new EST_Wave;
 
     if (di_db == 0)
     {
@@ -371,22 +371,18 @@ LISP FT_reslpc_resynth(LISP file)
 
     wfree(as);
 
-    // Add wave as stream into utterance
-    w->set_data_copy(output->track+di_db->sig_band,
-		     output->o_sz-(2*di_db->sig_band),
-		     di_db->samp_freq,
-		     1);
-
-    if (!u->stream_present("Wave"))
-	u->create_stream("Wave");
-
+    // Add wave into utterance
+    EST_Wave *w = new EST_Wave;
+    w->resize(output->o_sz-(2*di_db->sig_band),1);
+    for (int i=0; i<w->length(); i++)
+	w->a_no_check(i) = output->track[di_db->sig_band+i];
+    w->set_sample_rate(di_db->samp_freq);
     delete_output(output);
 
-    item.init("Wave");
-    item.set_contents(w,gc_wave);
-    u->stream("Wave").append(item);
+    item = u->create_relation("Wave")->append();
+    item->set_val("wave",est_val(w));
 
-    return siod_make_utt(*u);
+    return siod(u);
 }
 
 static void make_silence(DIPHONE_DATABASE *db,DIPHONE_SPN *ps,DIPHONE_OUTPUT *o)
@@ -413,9 +409,9 @@ static void make_silence(DIPHONE_DATABASE *db,DIPHONE_SPN *ps,DIPHONE_OUTPUT *o)
 static void merge_silences(EST_Utterance &u)
 {
     //  Remove multiple silences
-    EST_Stream_Item *s, *ns;
+    EST_Item *s, *ns;
     
-    for (s=u.stream("Segment").head(); s != 0;)
+    for (s=u.relation("Segment")->first(); s != 0; s=ns)
     {
 	ns = next(s);
 	if ((ns != 0) &&
@@ -423,10 +419,8 @@ static void merge_silences(EST_Utterance &u)
 	    (s->name() == ns->name())) // same *type* of silence
 	{
 	    // delete this segment
-	    u.remove_item("Segment",s->addr());
-	    
+	    remove_item(s,"Segment");
 	}
-	s = ns;
     }
 }
 
@@ -563,16 +557,16 @@ static DIPHONE_SPN *make_spn(EST_Utterance &u)
 {
     //  Build initial structure for donovan code
     DIPHONE_SPN *ps = walloc(DIPHONE_SPN,1);
-    const EST_Stream &seg = u.stream("Segment");
-    const EST_Stream &targ = u.stream("Target");
-    EST_Stream_Item *s;
+    EST_Relation *seg = u.relation("Segment");
+    EST_Relation *targ = u.relation("Target");
+    EST_Item *s;
     LISP cps;
     const char *ph_name;
-    EST_TBI *rt;
+    EST_Item *rt;
     int i;
-    float pos;
+    float pos,seg_start,seg_dur,seg_end;
 
-    ps->p_sz = seg.length();
+    ps->p_sz = seg->length();
     ps->p_max = ps->p_sz+1;
     ps->t_sz = 0;
     ps->phons = walloc(char *,ps->p_max);
@@ -586,28 +580,31 @@ static DIPHONE_SPN *make_spn(EST_Utterance &u)
     for (i=0; i<ps->p_sz; i++)
 	ps->diphs[i] = walloc(char,20);
 
-    ps->t_max = targ.length()+4;
+    ps->t_max = num_leaves(targ->head())+4;
     ps->pc_targs = walloc(int,ps->t_max);
     ps->targ_phon = walloc(int,ps->t_max);
     ps->targ_freq = walloc(int,ps->t_max);
     ps->abs_targ = walloc(int,ps->t_max);
     
     // Ensure there is a target at the start
-    if ((u.stream("Target").length() == 0) ||
-	(u.stream("Target").head()->end() != 0))
+    if ((targ->length() == 0) || (targ->first_leaf()->F("pos") != 0))
     {
 	ps->targ_phon[0] = 0;
-	if (u.stream("Target").length() == 0)
+	if (targ->length() == 0)
 	    ps->targ_freq[0] = di_db->def_f0;
 	else
-	    ps->targ_freq[0] = (int)atof(u.stream("Target").head()->name());
+	    ps->targ_freq[0] = targ->first()->I("f0");
 	ps->pc_targs[0] = 0;
 	ps->t_sz++;
     }
-    for (i=0,s=seg.head(); s != 0; s=next(s),i++)
+    seg_end = 0;
+    for (i=0,s=seg->first(); s != 0; s=next(s),i++)
     {
+	seg_start = seg_end;
+	seg_end = s->F("end");
+	seg_dur = seg_end - seg_start;
 	EST_String pph_name;  // possibly calculated phone name
-	if ((pph_name=s->feature("diphone_phone_name").string()) != "0")
+	if ((pph_name=s->f("diphone_phone_name").string()) != "0")
 	    ph_name = pph_name;
 	else if (((cps=ft_get_param("PhoneSet")) == NIL) ||
 	    ((streq(get_c_string(cps),di_db->phoneset))))
@@ -615,34 +612,31 @@ static DIPHONE_SPN *make_spn(EST_Utterance &u)
 	else
 	    ph_name = map_phone(s->name(),get_c_string(cps),di_db->phoneset);
 	ps->phons[i] = wstrdup(ph_name);
-	ps->duration[i] = (int)(s->dur() * di_db->samp_freq);  /* in frames */
+	ps->duration[i] = (int)(seg_dur * di_db->samp_freq);  /* in frames */
 	if (i>0)
 	    ps->cum_dur[i] = ps->cum_dur[i-1];
 	else
 	    ps->cum_dur[i] = 0;
 	ps->cum_dur[i] += ps->duration[i];
-	EST_Relation *target_rel = s->link("Target");
-	for (rt = target_rel->head();
+	for (rt = daughter1(s,"Target");
 	     rt != 0;
 	     rt = next(rt),ps->t_sz++)
 	{
-	    const EST_Stream_Item &t = u.ritem("Target", (*target_rel)(rt));
 	    ps->targ_phon[ps->t_sz] = i;
-	    ps->targ_freq[ps->t_sz] = atoi(t.name());
-	    pos = ((t.end() - s->start()) / s->dur()) * 99.9;
+	    ps->targ_freq[ps->t_sz] = rt->I("f0");
+	    pos = ((rt->F("pos") - seg_start) / seg_dur) * 99.9;
 	    ps->pc_targs[ps->t_sz] = (int)pos;
 	}
     }
     // Ensure there is a target at the end 
-    if ((u.stream("Target").length() == 0) ||
-	(u.stream("Target").tail()->end() != 
-	 u.stream("Segment").tail()->end()))
+    if ((targ->length() == 0) ||
+	(targ->last_leaf()->F("pos") != seg->last_leaf()->F("end")))
     {
 	ps->targ_phon[ps->t_sz] = i-1;
-	if (u.stream("Target").length() == 0)
+	if (targ->length() == 0)
 	    ps->targ_freq[ps->t_sz] = di_db->def_f0;
 	else
-	    ps->targ_freq[ps->t_sz] = (int)atof(u.stream("Target").tail()->name());
+	    ps->targ_freq[ps->t_sz] = targ->last_leaf()->I("f0");
 	ps->pc_targs[ps->t_sz] = 100;
 	ps->t_sz++;
     }
@@ -729,22 +723,24 @@ void di_add_diphonedb(DIPHONE_DATABASE *db)
     LISP lpair;
     DIPHONE_DATABASE *ddb;
 
+    if (diphone_dbs == NIL)
+	gc_protect(&diphone_dbs);
+	
     lpair = siod_assoc_str(db->name,diphone_dbs);
 
     if (lpair == NIL)
     {   // new diphone db of this name
 	diphone_dbs = cons(cons(rintern(db->name),
-				cons(MKPTR(db),NIL)),
+				cons(siod(db),NIL)),
 			   diphone_dbs);
-	gc_protect(&diphone_dbs);
     }
     else
     {	// already one of this name, don't know howto free it
 	cerr << "Diphone: warning redefining diphone database "
 	    << db->name << endl;
-	ddb = (DIPHONE_DATABASE *)(PTRVAL(car(cdr(lpair))));
+	ddb = diphone_db(car(cdr(lpair)));
 	delete_diphone_db(ddb);
-	PTRVAL(car(cdr(lpair))) = db;
+	setcar(cdr(lpair),siod(db));
     }
     
     di_db = db;
@@ -766,7 +762,7 @@ LISP FT_Diphone_select(LISP name)
     }
     else
     {
-	DIPHONE_DATABASE *db = (DIPHONE_DATABASE *)PTRVAL(car(cdr(lpair)));
+	DIPHONE_DATABASE *db = diphone_db(car(cdr(lpair)));
 	di_db = db;
     }
 
@@ -797,7 +793,7 @@ void festival_diphone_init(void)
 
     // New diphone synthesizer
     proclaim_module("diphone", &di_description);
-#ifdef HAVE_DI_PSOLA_TM
+#ifdef SUPPORT_PSOLA_TM
     // Only available for research
     proclaim_module("di_psolaTM");
 #endif
